from pytest import raises
from lyncs_mpi import Client


def test_array():
    client = Client(num_workers=4, launch=True)
    cart = client.comm.create_cart([2, 2, 1])

    for shape in (4, 4), (4, 2, 2), (2, 2, 2, 2):
        arr = cart.ones(shape)
        assert arr.sum() == 16

        futures = cart.get_futures(arr)
        assert len(futures) == len(cart)
        assert cart.workers == tuple(client.who_has(ftr)[0] for ftr in futures)

        arr = cart.array(futures)
        assert arr.sum() == 16
        assert tuple(arr.dask.keys()) == tuple(val.key for val in arr.dask.values())

    # Test errors
    with raises(ValueError):
        cart.ones((1, 2))

    with raises(TypeError):
        cart.get_futures(cart)

    with raises(ValueError):
        cart.array(futures[:1])

    cart2 = client.comm.create_cart([2, 1, 2])

    arr = cart2.ones((4, 2, 2))
    with raises(ValueError):
        cart.get_futures(arr)

    arr = cart.ones((4, 4))
    with raises(ValueError):
        cart2.get_futures(arr)

    with raises(ValueError):
        cart2.ones((4, 4))



"""
Utils for interfacing to MPI libraries using mpi4py and dask
"""

__version__ = "0.1.4"

from . import abc
from .lib import *
from .client import *
from .comm import *
from .distributed import *
from .cartesian import *
from .cart_array import *



"Tools for distributed classes on a Cartesian communicator"

__all__ = [
    "CommLocal",
    "Cartesian",
    "CartesianClass",
]

from functools import wraps
from itertools import chain
from dask.array import Array
from .comm import CartComm
from .distributed import Distributed, DistributedClass, Local


class Cartesian(Distributed):
    "Class for distributed objects on a Cartesian communicator"

    __slots__ = [
        "_comm",
    ]

    def __init__(self, ftrs, comm):
        assert isinstance(comm, CartComm)
        super().__init__(ftrs)
        self._comm = comm

    @property
    def comm(self):
        "Returns the communicator"
        return self._comm

    @property
    def coords(self):
        "Coordinates of the cartesian communicator"
        return self.comm.coords

    @property
    def procs(self):
        "Number of processes per dimension"
        return self.comm.dims

    @property
    def ranks(self):
        "Ranks of the communicator"
        return self.comm.ranks

    @classmethod
    def _remote_call(cls, *args, **kwargs):
        # Looking for cartesian comm
        comms = None
        for arg in chain(args, kwargs.values()):
            if isinstance(arg, Cartesian):
                arg = arg.comm
            if isinstance(arg, CartComm):
                if comms is not None and comms != arg:
                    raise ValueError(
                        "Different cartesian communicator passed to class init"
                    )
                comms = arg
        if comms is None:
            raise ValueError(
                "Cartesian communicator not found in class init in parallel mode"
            )

        # Looking for dask arrays
        get_arrays = (
            lambda arr: comms.get_futures(arr) if isinstance(arr, Array) else arr
        )
        args = (get_arrays(arg) for arg in args)
        kwargs = {key: get_arrays(arg) for key, arg in kwargs.items()}

        return Cartesian(super()._remote_call(*args, **kwargs), comms)

    @wraps(CartComm.index)
    def index(self, key):
        return self.comm.index(key)


class CommLocal(Local):
    "Mock class of Cartesian"

    @property
    def comm(self):
        "Mock of Cartesian.comm"
        return None

    @property
    def coords(self):
        "Mock of Cartesian.coords"
        return ((0,),)

    @property
    def procs(self):
        "Mock of Cartesian.procs"
        return (1,)

    @property
    def ranks(self):
        "Mock of Cartesian.ranks"
        return (0,)

    def index(self, key):
        "Mock of Cartesian.index"
        if key == 0:
            return 0
        if isinstance(key, tuple) and all((_ == 0 for _ in key)):
            return 0
        return super().index(key)


class CartesianClass(DistributedClass):
    "Metaclass for cartesian distributed classes"

    @staticmethod
    def local_class():
        return CommLocal

    @staticmethod
    def distributed_class():
        return Cartesian



from dask.base import wait
from lyncs_mpi import Client
from lyncs_mpi.testing import CartesianTest

client = Client(2)
comm = client.create_comm().create_cart((2,))


def test_bench_init(benchmark):
    benchmark(CartesianTest, (4, 4), comm=comm)


def test_bench_result(benchmark):
    test = CartesianTest((4, 4), comm=comm)
    benchmark(test.range, 10)


def test_bench_array(benchmark):
    test = CartesianTest((4, 4), comm=comm)
    benchmark(test.ones)


def test_bench_array_wait(benchmark):
    test = CartesianTest((4, 4), comm=comm)
    benchmark(lambda: wait(test.ones()))



"Cartesian arrays, arrays defined over a cartesian communicator"

__all__ = [
    "CartComm",
]

from math import ceil
from dask.core import flatten
from dask.array import Array, zeros, ones, empty, full
from dask.array.core import normalize_chunks
from lyncs_utils import add_to
from .comm import CartComm


class KeyPatch(tuple):
    "This is a patch for accepting a tuple as a dask key"

    def __dask_graph__(self):
        return {self: None}

    def __dask_keys__(self):
        return [self]


@add_to(CartComm)
def check_dims(self, nchunks):
    "Checks the number of chunks of an array versus the communicator dimensions"
    dims = self.dims
    if len(dims) < len(nchunks):
        dims += (1,) * (len(nchunks) - len(dims))
    elif len(dims) > len(nchunks):
        if not all((dim == 1 for dim in dims[len(nchunks) :])):
            raise ValueError("Array shape smaller than distributed axes")
        dims = dims[: len(nchunks)]
    if nchunks != dims:
        raise ValueError(
            f"""Number of chunks not compatible with comm dims.
        Got nchunks = {nchunks}, while expected {dims}."""
        )


@add_to(CartComm)
def get_futures(self, arr):
    """
    Return the list of futures of a Dask array associated to a cartesian communicator.

    Parameters
    ----------
    arr: Dask Array
        A dask array distributed as the cartesian communicator. Note: the array can
        have more axes than the communicator as long as they are not distributed.
    """
    if not isinstance(arr, Array):
        raise TypeError(f"Expected a Dask Array; got {type(arr)}.")

    self.check_dims(tuple(len(chunks) for chunks in arr.chunks))

    idxs, _ = zip(*self.normalize_dims())
    coords = self.normalize_coords()
    keys = tuple(flatten(arr.__dask_keys__()))
    key_idx = {}
    for key in keys:
        coord = tuple(key[_i + 1] for _i in idxs)
        key_idx[key] = coords.index(coord)

    keys = sorted(keys, key=key_idx.__getitem__)
    restrictions = {KeyPatch(key): worker for key, worker in zip(keys, self.workers)}

    arr = arr.persist(workers=restrictions)
    assert len(self) == len(arr.dask.values())

    return list(arr.dask[key] for key in keys)


@add_to(CartComm)
def array(self, futures, shape=None, chunks=None, dtype=None):
    """
    Turns a set of future arrays (result of a distributed operation),
    associated to a cartesian communicator, into a Dask Array.

    Parameters
    ----------
    cart: CartComm
        A cartesian communicator with dimensions equal to the number of chunks
    futures: tuple(futures)
        A set of future arrays associated to the cart
    dims_axes: tuple
        The axes associated to the dimensions of the cart
    shape: tuple(int)
        The shape of the array
    chunks: tuple(int)
        The chunks of the array
    dtype: tuple(int)
        The dtype of the array
    """
    if not len(futures) == len(self):
        raise ValueError("futures and cart must have the same length")

    if chunks is None or dtype is None:
        infos = self.client.map(lambda arr: (arr.dtype, arr.shape), futures)
        infos = tuple(_.result() for _ in infos)

        if dtype is None:
            dtype = infos[0][0]
        if not all((dtype == dtp for (dtp, _) in infos)):
            raise TypeError(
                f"Futures have different dtypes {[info[0] for info in infos]}"
            )

        if chunks is None:
            chunks = infos[0][1]
            if not all((chunks == chn for (_, chn) in infos)):
                # TODO: normalize chunks using shape
                raise NotImplementedError(
                    "Futures with non-uniform chunks not supported yet"
                )

    if shape is None:
        shape = list(chunks)
        for _i, _l in self.normalize_dims():
            shape[_i] *= _l

    chunks = normalize_chunks(chunks, shape, dtype=dtype)

    self.check_dims(tuple(len(chunk) for chunk in chunks))

    dask = {}
    idxs, _ = zip(*self.normalize_dims())
    for coords, future in zip(self.normalize_coords(), futures):
        key = [0] * len(shape)
        for _i, _c in zip(idxs, coords):
            key[_i] = _c

        name = next(iter(futures)).key
        if isinstance(name, tuple):
            name = name[0]
        assert isinstance(name, str)
        key = (name,) + tuple(key)
        dask[key] = future

    return Array(dask, next(iter(dask.keys()))[0], chunks, dtype=dtype, shape=shape)


@add_to(CartComm)
def get_chunks(self, shape):
    "Returns the chunks of shape compatible with the CartComm"
    chunks = list(shape)
    for _i, _l in self.normalize_dims():
        if _i >= len(shape):
            raise ValueError("Array shape smaller than distributed axes")
        if _l > shape[_i]:
            raise ValueError(f"Cannot chunk shape on axis {_i} in {_l} pieces")
        chunks[_i] = ceil(chunks[_i] / _l)
    return chunks


def array_wrapper(mth):
    """
    Cartesian variant of %s (documentation follows).
    Assigns to the workers of a cartesian communicator chunks of the array.

    The first argument `cart` should be a cartesian communicator
    with dimensions equal to the number of chunks of the array.

    Additional option is `dims_axes` that can be used to reorder
    the dimensions of the communicator to match the chunked axes.

    --------- END ----------
    """

    def wrapper(self, shape, **kwargs):
        chunks = self.get_chunks(shape)
        arr = mth(shape, chunks=chunks, **kwargs)

        idxs, _ = zip(*self.normalize_dims())
        coords = self.normalize_coords()
        workers = self.workers
        keys = flatten(arr.__dask_keys__())
        restrictions = {}
        for key in keys:
            coord = tuple(key[_i + 1] for _i in idxs)
            restrictions[KeyPatch(key)] = workers[coords.index(coord)]

        return arr.persist(workers=restrictions)

    wrapper.__name__ = "cart_" + mth.__name__
    wrapper.__doc__ = array_wrapper.__doc__ % mth.__name__ + mth.__doc__
    return wrapper


CartComm.zeros = array_wrapper(zeros)
CartComm.ones = array_wrapper(ones)
CartComm.empty = array_wrapper(empty)
CartComm.full = array_wrapper(full)



"""
Some ABC to be used for typing parallel classes
"""
# pylint: disable=too-few-public-methods

__all__ = [
    "Result",
    "Global",
    "Array",
]

from abc import ABC
from functools import partial
from numpy import ndarray
from dask.array import Array as daskArray
from dask.array import from_array
from .distributed import results, Distributed
from .cartesian import Cartesian


class Result(ABC):
    "Returns a tuple with the result of the remote calls"

    def __new__(cls, ftrs, **kwargs):
        return results(*ftrs)


class Global(Result):
    "Returns a single value supposed to be the same result for all the remote calls"

    def __new__(cls, ftrs, **kwargs):
        res = super().__new__(cls, ftrs)
        ret = res[0]
        if not all((_ == ret for _ in res)):
            raise RuntimeError(f"Expected global value but got different resuts: {res}")
        return ret


class Constant(Global):
    "Like Global, but the value is stored for future use and avoiding remote call"

    def __new__(cls, ftrs, caller=None, key=None, **kwargs):
        assert key, "Expected a key"
        assert caller and isinstance(
            caller, Distributed
        ), "Expected a Distributed caller. Got {caller}"
        ret = super().__new__(cls, ftrs, caller=caller, key=key, **kwargs)
        caller._constants[key] = ret
        return ret


class Array(Result):
    "Returns a dask array supposing the futures output to be numpy-like arrays"

    def __new__(cls, *args, **kwargs):
        if not args:
            return partial(cls, **kwargs)

        assert len(args) == 1
        arg = args[0]

        array_kwargs = {}
        for key in ("shape", "chunks", "dtype"):
            if key not in kwargs:
                continue
            val = kwargs[key]
            if callable(val):
                assert "caller" in kwargs
                val = val(kwargs["caller"])
            array_kwargs[key] = val

        if isinstance(arg, Cartesian):
            return arg.comm.array(arg, **array_kwargs)

        if not isinstance(arg, (ndarray, daskArray)):
            raise TypeError(f"Unexpected type {type(arg)}")

        shape = array_kwargs.get("shape", arg.shape)
        if arg.shape != shape:
            raise ValueError(
                f"Not compatible shape. Got {arg.shape} but expected {shape}"
            )
        chunks = array_kwargs.get("chunks", kwargs["caller"].comm.get_chunks(shape))

        if isinstance(arg, ndarray):
            return from_array(arg, chunks=chunks)

        if arg.chunksize != chunks:
            return arg.rechunk(chunks)
        return arg


Array.register(ndarray)
Array.register(daskArray)



"""
Just some simple classes used by the tests.
(They are needed here to avoid pickling issues)
"""

from random import random
import numpy as np
from .abc import Result, Global, Constant, Array
from .distributed import DistributedClass
from .cartesian import CartesianClass


class DistributedTest(metaclass=DistributedClass):
    "A distributed test class"

    def __init__(self, value=None):
        self.value = value

    @property
    def ten(self) -> Constant:
        "Returns 10"
        return 10

    def range(self, length) -> Global:
        "Returns range(length)"
        return range(length)

    def not_global(self) -> Global:
        "Function for error testing"
        return random()

    def values(self) -> Result:
        "Returns the values"
        return self.value


class CartesianTest(DistributedTest, metaclass=CartesianClass):
    "A cartesian distributed test class"

    Field = Array(
        shape=lambda self: self.shape,
        chunks=lambda self: self.lshape,
        dtype=lambda self: self.dtype,
    )

    def __init__(self, lshape, value=None, dtype="int", comm=None):
        super().__init__(value)
        self._lshape = lshape
        if comm:
            procs = comm.dims
            if len(procs) < len(lshape):
                procs += (1,) * (len(lshape) - len(procs))
            elif len(procs) > len(lshape):
                assert all((i == 1 for i in procs[len(lshape) :]))
            self._shape = tuple(i * j for i, j in zip(lshape, procs))
        else:
            self._shape = lshape
        self._dtype = dtype

    @property
    def lshape(self) -> Constant:
        "Local shape"
        return self._lshape

    @property
    def shape(self) -> Constant:
        "Global shape"
        return self._shape

    @property
    def dtype(self) -> Constant:
        "Array dtype"
        return self._dtype

    def ones(self) -> Field:
        "Returns an array of ones"
        return np.ones(self.lshape, self.dtype)

    def mul_by_value(self, arr: Field) -> Array:
        "Multiplies the array by value"
        return arr * self.value



from pytest import raises
from lyncs_mpi import (
    Client,
    default_client,
)


def test_comm():
    client = Client(num_workers=4, launch=True)
    comm = client.comm
    assert len(comm) == len(client.workers)
    assert len(comm) == comm.size
    assert set(comm.ranks) == set(range(4))
    assert set(comm.workers) == set(client.workers)
    assert comm[0] == comm[comm.ranks_workers[0]]

    with raises(KeyError):
        comm[5]

    cart = comm.create_cart([2, 2], periods=True)
    assert cart.dims == (2, 2)
    assert all(cart.periods)
    assert cart[0] == cart[cart.ranks_coords[0]]

    comm1 = client.create_comm(2)
    comm2 = client.create_comm(2, exclude=comm1.workers)
    assert not set(comm1.workers).intersection(comm2.workers)
    assert set(comm1.workers + comm2.workers) == set(client.workers)



"Distributed MPI communicators"

__all__ = [
    "Comm",
    "CartComm",
]

from lyncs_utils import compute_property
from .distributed import Distributed, results


class Comm(Distributed):
    "MPI communicator"

    __slots__ = [
        "_ranks",
    ]

    def __init__(self, comms):
        super().__init__(comms, cls=self.type)

    @property
    def type(self):
        "Returns the MPI type of the class"
        # pylint: disable=import-outside-toplevel
        from mpi4py import MPI

        return MPI.Comm

    @property
    def size(self):
        "Size of the communicator"
        return len(self)

    @compute_property
    def ranks(self):
        "Ranks of the communicator with respective worker"
        return results(*self.rank)

    @property
    def ranks_workers(self):
        "Mapping between ranks and workers"
        return dict(zip(self.ranks, self.workers))

    def create_cart(self, dims, periods=True, reorder=False):
        """
        Makes a new communicator to which topology information has been attached

        Parameters
        ----------
        dims: list
            integer array specifying the number of processes in each dimension
        periods: boolean [list]
            logical [array] specifying whether the grid is periodic (True) or not (False)
        reorder: boolean
            ranking may be reordered (True) or not (False)
        """
        return CartComm(self.Create_cart(dims, periods=periods, reorder=reorder))

    def index(self, key):
        "Returns the index of key that can be either rank(int) or worker(str)"
        if isinstance(key, int) and key in self.ranks:
            return self.ranks.index(key)
        if isinstance(key, str) and key in self.workers:
            return self.workers.index(key)
        raise KeyError(f"{key} is neither a rank or a worker of {self}")


class CartComm(Comm):
    "Cartesian communicator"

    __slots__ = [
        "_dims",
        "_periods",
        "_coords",
    ]

    def __init__(self, comms):
        super().__init__(comms)
        topos = results(*self.Get_topo())
        self._dims = tuple(topos[0][0])
        self._periods = tuple(bool(_) for _ in topos[0][1])
        self._coords = tuple(tuple(topo[2]) for topo in topos)

    @property
    def type(self):
        "Returns the MPI type of the class"
        # pylint: disable=import-outside-toplevel
        from mpi4py import MPI

        return MPI.Cartcomm

    @property
    def dims(self):
        "Dimensions of the cartesian communicator"
        return self._dims

    @property
    def periods(self):
        "Periodicity of the cartesian communicator"
        return self._periods

    @property
    def coords(self):
        "Coordinates of the cartesian communicator"
        return self._coords

    @property
    def ranks_coords(self):
        "Coordinates of the ranks of the cartesian communicator"
        return dict(zip(self.ranks, self.coords))

    def index(self, key):
        "Returns the index of key that can be either rank(int) or worker(str) or coord(tuple)"
        if isinstance(key, tuple):
            lkey = len(key)
            ldims = len(self.dims)
            if len(key) <= len(self.dims):
                key = key + (0,) * (ldims - lkey)
            elif key[ldims:] == (0,) * (lkey - ldims):
                key = key[:ldims]
            else:
                raise KeyError(f"{key} out of range {self.dims}")
            return self.coords.index(key)
        return super().index(key)

    def normalize_dims(self):
        "Removes non-distributed dimensions from dims. Returns tuple (index, size)"
        return tuple((_i, _l) for _i, _l in enumerate(self.dims) if _l > 1)

    def normalize_coords(self):
        "Removes non-distributed dimensions from coords."
        dims = self.normalize_dims()
        return tuple(tuple(coord[_i] for _i, _ in dims) for coord in self.coords)

    # NOTE: additional methods are implemented in cart_array.py



from pytest import raises
from lyncs_mpi import Client, CartComm
from lyncs_mpi.distributed import *
from lyncs_mpi.cartesian import *
from lyncs_mpi.testing import DistributedTest, CartesianTest
from dask.array import Array, ones
from numpy import ndarray
import numpy as np


def test_commlocal():
    test = CartesianTest((4, 2), value=10)
    assert test.value == 10
    assert isinstance(test, CommLocal)
    assert isinstance(test, Local)
    assert isinstance(test, CartesianTest)
    assert isinstance(test, DistributedTest)
    assert len(test) == 1
    assert test.client == None
    assert test.workers == ("localhost",)
    assert test in test

    assert test.comm is None
    assert test.coords == ((0,),)
    assert test.procs == (1,)
    assert test.ranks == (0,)

    assert test[0] == test["localhost"]
    assert test[0] == test[0, 0, 0, 0, 0]

    arr = test.ones()
    assert arr.shape == (4, 2)
    assert isinstance(arr, ndarray)
    assert arr.sum() == 8


def test_cartesian():
    client = Client(2)
    cart = client.create_comm().create_cart((2,))
    assert len(cart) == 2

    test = CartesianTest((2, 2), value=1, comm=cart)
    assert isinstance(test, Cartesian)
    assert isinstance(test, Distributed)
    assert isinstance(test, CartesianTest)
    assert isinstance(test, DistributedTest)
    assert test.client is client
    assert len(test) == 2
    assert set(test.workers) == set(cart.workers)

    assert test.ten == 10
    assert test.values() == (1, 1)

    assert test.comm is cart
    assert test.coords == ((0,), (1,))
    assert test.procs == (2,)
    assert test.ranks == (0, 1)

    assert test[0] == test[0, 0]
    assert test[1] == test[1, 0]

    assert isinstance(test.comm, CartComm)
    arr = test.ones()
    assert arr.shape == (4, 2)
    assert isinstance(arr, Array)
    assert arr.sum() == 8

    test.value = 10
    assert (test.mul_by_value(arr) == arr * 10).all()

    np_arr = np.ones(arr.shape)
    assert (test.mul_by_value(np_arr) == np_arr * 10).all()

    da_arr = ones(arr.shape, chunks=(1, 2))
    assert (test.mul_by_value(np_arr) == np_arr * 10).all()

    da_arr = ones(arr.shape, chunks=arr.chunksize)
    assert (test.mul_by_value(np_arr) == np_arr * 10).all()

    with raises(TypeError):
        test.mul_by_value(1)

    with raises(ValueError):
        np_arr = np.ones((1, 1))
        test.mul_by_value(np_arr)

    with raises(ValueError):
        cart2 = client.create_comm(1).create_cart((1,))
        test = CartesianTest(cart2, comm=cart)

    with raises(ValueError):
        init = client.scatter((1, 2))
        test = CartesianTest(init)



"""
This file is automatically generated during the setup.
Please do not edit config.py, but edit config.py.in in the root directory.
"""

MPI_INCLUDE_DIRS = "/usr/lib/x86_64-linux-gnu/openmpi/include/openmpi;/usr/lib/x86_64-linux-gnu/openmpi/include/openmpi/opal/mca/event/libevent2022/libevent;/usr/lib/x86_64-linux-gnu/openmpi/include/openmpi/opal/mca/event/libevent2022/libevent/include;/usr/lib/x86_64-linux-gnu/openmpi/include"
MPI_LINK_FLAGS = "-L/usr/lib -pthread"
MPI_LIBRARIES = "/usr/lib/x86_64-linux-gnu/openmpi/lib/libmpi.so"



from pytest import raises
from lyncs_mpi import Client
from lyncs_mpi.distributed import *
from lyncs_mpi.testing import DistributedTest


def test_local():
    test = DistributedTest(10)
    assert test.value == 10
    assert isinstance(test, Local)
    assert isinstance(test, DistributedTest)
    assert len(test) == 1
    assert test.client == None
    assert test.workers == ("localhost",)
    assert test in test
    assert test["localhost"] is test
    assert test.dask is None
    assert test.type is DistributedTest
    assert test.wait() is test

    assert test.ten == 10
    assert test.range(5) == range(5)

    with raises(KeyError):
        test["bar"]

    with raises(ValueError):
        Distributed._remote_call(int, 1)

    with raises(TypeError):
        Distributed._remote_call(1)

    assert Distributed._set_and_return(test, "value", 1) is test
    assert test.value == 1

    assert Distributed._insert_args((), (), 1, 2, 3) == (1, 2, 3)
    assert Distributed._insert_args((0, 2), (1, 3), 2) == (1, 2, 3)
    assert test.value == 1


def test_distributed():
    client = Client(2)
    init = client.scatter((1, 2))
    assert len(init) == 2
    assert len(set(client.who_has(init))) == 2

    test = DistributedTest(init)
    assert isinstance(test, Distributed)
    assert isinstance(test, DistributedTest)
    assert test.client is client
    assert len(test) == 2
    assert set(test.workers) == set(client.who_has(init))
    # assert set(test.workers) == set(client.who_has(test))
    assert test.ten == 10
    assert test.values() == (1, 2)

    test = DistributedTest(value=init)
    assert isinstance(test, Distributed)
    assert isinstance(test, DistributedTest)
    assert test.client is client
    assert set(test.workers) == set(client.who_has(init))
    assert test.ten == 10
    assert "ten" in test._constants
    assert test.values() == (1, 2)

    assert test.range(5) == range(5)

    assert test.value.type == int

    test.value = "bar"
    assert test.value.type == str

    with raises(TypeError):
        test(1).wait()

    assert test[test.workers[0]] is next(iter(test))

    with raises(KeyError):
        test["bar"]

    with raises(RuntimeError):
        test.not_global()



import os
import sys
import sh
import tempfile
from pytest import raises
from lyncs_mpi import (
    Client,
    default_client,
)


def test_client():
    client = Client(num_workers=1)
    assert client._server is not None
    assert default_client() is client
    assert len(client.workers) == 1

    with raises(RuntimeError):
        client.select_workers(2)

    with raises(RuntimeError):
        client.select_workers(1, exclude=client.workers)

    with raises(NotImplementedError):
        client.select_workers(1, resources="GPU")

    client.close_server()
    assert client._server is None

    with raises(RuntimeError):
        client.close_server()

    client = Client(num_workers=1)
    client.__del__()


def test_not_launch():
    with raises(RuntimeError):
        Client(launch=False)

    pwd = os.getcwd()
    sh.cd(tempfile.mkdtemp())
    test = sh.mpirun(
        "-n",
        3,
        sys.executable,
        "-c",
        "from lyncs_mpi import Client; Client(1, launch=False)",
    )
    sh.cd(pwd)
    assert test.exit_code == 0



from lyncs_setuptools import setup, CMakeExtension

requirements = [
    "lyncs-setuptools",
    "mpi4py",
    "lyncs-cppyy",
    "lyncs-utils",
    "dask",
    "distributed",
    "dask[array]",
    "dask[distributed]",
    "dask-mpi",
    "sh>=1.14.0",
]


setup(
    "lyncs_mpi",
    ext_modules=[CMakeExtension("lyncs_mpi.lib", ".")],
    exclude=["*.config"],
    data_files=[(".", ["config.py.in"])],
    install_requires=requirements,
    extras_require={"test": ["pytest", "pytest-cov", "pytest-benchmark"]},
)



"Tools for dask distributed classes"

__all__ = [
    "isdistributed",
    "anydistributed",
    "Distributed",
    "Local",
    "DistributedClass",
    "results",
]

import uuid
from itertools import chain
from functools import wraps
from inspect import isclass
from distributed.client import Future
from distributed import as_completed
from lyncs_utils import (
    isiterable,
    interactive,
    compute_property,
    apply_annotations,
    select_kwargs,
)


def isdistributed(val):
    "Returns if the argument is a Dask distributed object"
    return isiterable(val, Future)


def anydistributed(*args, **kwargs):
    "Returns if any of the arguments is a Dask distributed object"
    return any((isdistributed(val) for val in chain(args, kwargs.values())))


def results(*args):
    "Returns the results of the futures in args"
    return tuple(arg.result() if isinstance(arg, Future) else arg for arg in args)


class Distributed:
    "Class for handling distributed objects"

    __slots__ = [
        "_dask",
        "_workers",
        "_type",
        "_constants",
        "_keys",
    ]

    def __init__(self, dask, cls=None):
        self._dask = tuple(dask)
        self._constants = dict()
        if cls:
            self._type = cls

    def wait(self):
        "Waits for all the distributed futures to be completed and raises errors if any"
        for ftr in as_completed(self.dask):
            if ftr.status == "error":
                ftr.result()
        return self

    @property
    def dask(self):
        "Returns the low-level dask objects"
        return self._dask

    def __dask_graph__(self):
        return dict(zip(self.__dask_keys__(), self.dask))

    def __dask_keys__(self):
        try:
            return self._keys
        except AttributeError:
            self._keys = tuple(str(uuid.uuid4()) for _ in range(len(self)))
            return self._keys

    @property
    def client(self):
        "Returns the Client managing the distributed objects"
        return next(iter(self)).client

    @compute_property
    def workers(self):
        "Returns the of workers holding the distributed objects"
        return tuple(map(next, map(iter, map(self.client.who_has, self.wait()))))

    @compute_property
    def type(self):
        "Returns the class of the distributed objects"
        return next(iter(self.wait())).type

    def index(self, key):
        "Returns the index of the dask futures (self.dask) for a given key"
        if isinstance(key, str) and key in self.workers:
            return self.workers.index(key)
        raise KeyError(f"Key {key} not found")

    def __iter__(self):
        return iter(self.dask)

    def __len__(self):
        return len(self.dask)

    def __getitem__(self, key):
        return self.dask[self.index(key)]

    @classmethod
    def _get_finalize(cls, fnc):
        if hasattr(fnc, "__annotations__"):
            ret = fnc.__annotations__.get("return", None)
            if callable(ret):
                return ret
        if isinstance(fnc, property):
            return cls._get_finalize(fnc.fget)
        return lambda _, **__: _

    def _finalize(self, fnc, value, **kwargs):
        try:
            return fnc(value, caller=self, **kwargs)
        except TypeError:
            return fnc(value)

    @staticmethod
    def _insert_args(idxs, vals, *args):
        "Auxiliary function that inserts vals into args at idxs position used by _remote_call"
        if not idxs:
            return args
        args = list(args)
        for i, arg in zip(idxs, vals):
            args.insert(i, arg)
        return tuple(args)

    @classmethod
    def _remote_call(cls, fnc, *args, **kwargs):
        if isdistributed(fnc):
            call = lambda fnc, *args, **kwargs: fnc(*args, **kwargs)
            args = (fnc,) + args
        else:
            if not callable(fnc):
                raise TypeError(f"Type {type(fnc)} is not callable")
            call = fnc

        keys = []
        vals = []
        for i, arg in enumerate(args):
            if isdistributed(arg):
                keys.append(i)
                vals.append(arg)

        args = tuple(arg for i, arg in enumerate(args) if i not in keys)
        n_args = len(keys)

        for key, arg in list(kwargs.items()):
            if isdistributed(arg):
                keys.append(key)
                vals.append(arg)
                kwargs.pop(key)

        if not vals:
            raise ValueError("No distributed argument found when calling fnc")

        client = next(iter(vals[0])).client

        return Distributed(
            client.map(
                lambda *vals: call(
                    *Distributed._insert_args(keys[:n_args], vals[:n_args], *args),
                    **kwargs,
                    **dict(zip(keys[n_args:], vals[n_args:])),
                ),
                *vals,
                pure=False,  # TODO: add support for pure functions
            ),
            cls=call if isclass(call) else None,
        )

    def __callattr__(self, key, *args, **kwargs):
        fnc = getattr(self.type, key)
        caller = lambda fnc, arg: select_kwargs(fnc, arg, caller=self, key=key)
        args, kwargs = apply_annotations(fnc, self, *args, _caller=caller, **kwargs)
        finalize = self._get_finalize(fnc)
        return caller(finalize, self._remote_call(fnc, *args, **kwargs))

    def __getattr__(self, key):
        if key in self._constants:
            return self._constants[key]
        if key in dir(type(self)):
            return getattr(type(self), key).__get__(self)
        try:
            attr = getattr(self.type, key)
            finalize = self._get_finalize(attr)
            if callable(attr):
                fnc = lambda *args, **kwargs: self.__callattr__(key, *args, **kwargs)
                if interactive():
                    return wraps(attr)(fnc)
                return fnc
            return select_kwargs(
                finalize, self._remote_call(getattr, self, key), key=key, caller=self
            )
        except AttributeError:
            pass
        return self._remote_call(getattr, self, key)

    def _set_and_return(self, key, val):
        "Auxiliary function used by __setattr__"
        setattr(self, key, val)
        return self

    def __setattr__(self, key, val):
        if key in dir(type(self)):
            super().__setattr__(key, val)
        else:
            self._dask = self._remote_call(
                Distributed._set_and_return, self, key, val
            ).dask

    def __call__(self, *args, **kwargs):
        return self._remote_call(self, *args, **kwargs)

    def __eq__(self, other):
        if isinstance(other, Distributed):
            other = other.dask
        return self.dask == other


class Local:
    "Mock class for enabling functions of Distributed into a local class"

    @property
    def dask(self):
        "Mock of Distributed.dask"
        return None

    @property
    def client(self):
        "Mock of Distributed.client"
        return None

    @property
    def workers(self):
        "Mock of Distributed.workers"
        return ("localhost",)

    @property
    def type(self):
        "Mock of Distributed.type"
        return type(self)

    def wait(self):
        "Mock of Distributed.wait"
        return self

    def index(self, key):
        "Mock of Distributed.index"
        if key in self.workers:
            return 0
        raise KeyError(f"Key {key} not found")

    def __len__(self):
        return 1

    def __iter__(self):
        yield self

    def __getitem__(self, key):
        assert self.index(key) == 0
        return self


class DistributedClass(type):
    "Metaclass for Dask distributed classes"

    @staticmethod
    def local_class():
        "The local class to use"
        return Local

    @staticmethod
    def distributed_class():
        "The distributed class to use"
        return Distributed

    def __new__(cls, name, bases, class_attrs, **kwargs):
        "Checks that cls does not have methods of Local and adds Local to bases"
        local = cls.local_class()
        if local not in bases:
            bases = (local,) + bases
        return super().__new__(cls, name, bases, class_attrs, **kwargs)

    def __call__(cls, *args, **kwargs):
        if anydistributed(*args, **kwargs):
            return type(cls).distributed_class()._remote_call(cls, *args, **kwargs)
        return super().__call__(*args, **kwargs)

    def __instancecheck__(cls, instance):
        return super().__instancecheck__(instance) or (
            isdistributed(instance) and issubclass(Distributed(instance).type, cls)
        )



"""
Specialized dask Client for MPI communicators
"""

__all__ = [
    "default_client",
    "Client",
]

import os
import sys
import time
import shutil
import signal
import atexit
import tempfile
import multiprocessing
from functools import wraps
import sh
from dask_mpi import initialize
from dask.distributed import Client as _Client
from dask.distributed import default_client as _default_client
from .lib import default_comm
from .comm import Comm


@wraps(_default_client)
def default_client():
    "Returns the default client"
    client = _default_client()
    assert isinstance(client, Client), "No MPI client found"
    return client


class Client(_Client):
    """
    Subclass of dask.distributed.Client specialized for MPI communicators
    The initialization follows the guidelines of http://mpi.dask.org/
    automatizing the process of creating MPI-distributed dask workers.
    """

    def __init__(
        self, num_workers=None, threads_per_worker=1, launch=None, out=None, err=None
    ):
        """
        Returns a Client connected to a cluster of `num_workers` workers.
        """
        self._server = None

        if launch is None:
            launch = default_comm().size == 1

        # pylint: disable=import-outside-toplevel,
        if not launch:
            # Then the script has been submitted in parallel with mpirun
            num_workers = num_workers or default_comm().size - 2
            if num_workers < 0 or default_comm().size != num_workers + 2:
                raise RuntimeError(
                    f"""
                Error: (num_workers + 2) processes required.
                The script has not been submitted on enough processes.
                Got {default_comm().size} processes instead of {num_workers + 2}.
                """
                )

            initialize(
                nthreads=threads_per_worker,
                local_directory=tempfile.mkdtemp(),
                nanny=False,
            )

            super().__init__()

        else:
            num_workers = num_workers or (multiprocessing.cpu_count() + 1)

            # Since dask-mpi produces several file we create a temporary directory
            self._dir = tempfile.mkdtemp()

            # The command runs in the background (_bg=True)
            # and the stdout(err) is stored in self._out(err)
            pwd = os.getcwd()
            sh.cd(self._dir)
            self._server = sh.mpirun(
                "-n",
                num_workers + 1,
                "dask-mpi",
                "--no-nanny",
                "--nthreads",
                threads_per_worker,
                "--scheduler-file",
                "scheduler.json",
                _bg=True,
                _out=out or sys.stdout,
                _err=err or sys.stderr,
            )
            sh.cd(pwd)

            atexit.register(self.close_server)

            super().__init__(scheduler_file=self._dir + "/scheduler.json")

        # Waiting for all the workers to connect
        def handler(signum, frame):
            if self.server is not None:
                self.close_server()
            raise RuntimeError(
                "Couldn't connect to %d processes. Got %d workers."
                % (num_workers, len(self.workers))
            )

        signal.signal(signal.SIGALRM, handler)
        signal.alarm(5)

        while len(self.workers) != num_workers:
            time.sleep(0.001)

        signal.alarm(0)

        self.ranks = {key: val["name"] for key, val in self.workers.items()}
        self._comm = self.create_comm()

    @property
    def comm(self):
        "Returns the global communicator of the clients"
        return self._comm

    @property
    def workers(self):
        "Returns the list of workers."
        return self.scheduler_info()["workers"]

    @property
    def server(self):
        "Returns the running server if available"
        return self._server

    def close_server(self):
        """
        Closes the running server
        """
        if self.server is None:
            raise RuntimeError("No MPI-server started by the client")
        self.shutdown()
        self.close()
        self.server.wait()
        shutil.rmtree(self._dir)
        self._server = None
        atexit.unregister(self.close_server)

    def __del__(self):
        """
        In case of server started, closes the server
        """
        if self.server is not None:
            self.close_server()

        if hasattr(self, "_timeout"):
            super().__del__()

    def who_has(self, *args, overload=True, **kwargs):
        """
        Overloading of distributed.Client who_has.
        Checks that only one worker owns the futures and returns the list of workers.

        Parameters
        ----------
        overload: bool, default true
            If false the original who_has is used
        """
        if overload:
            _workers = list(super().who_has(*args, **kwargs).values())
            workers = [w[0] for w in _workers if len(w) == 1]
            assert len(workers) == len(
                _workers
            ), "More than one process has the same reference"
            return workers

        return super().who_has(*args, **kwargs)

    def select_workers(
        self, num_workers=None, workers=None, exclude=None, resources=None
    ):
        """
        Selects `num_workers` from the one available.

        Parameters
        ----------
        workers: list, default all
          List of workers to choose from.
        exclude: list, default none
            List of workers to exclude from the total.
        resources: dict, default none
            Defines the resources the workers should have.
        """

        if not workers:
            workers = list(self.ranks.keys())

        workers = set(workers)
        workers = workers.intersection(self.ranks.keys())

        if exclude:
            if isinstance(exclude, str):
                exclude = [exclude]
            workers = workers.difference(exclude)

        if resources:
            # TODO select accordingly requested resources
            raise NotImplementedError("Resources not implemented.")

        if not num_workers:
            num_workers = len(workers)

        if num_workers > len(workers):
            raise RuntimeError("Available workers are less than required")

        # TODO implement some rules to choose wisely n workers
        # e.g. workers less busy, close to each other, etc
        selected = list(workers)[:num_workers]

        return selected

    @wraps(select_workers)
    def create_comm(self, *args, **kwargs):
        """
        Return a MPI communicator involving workers available by the client.

        Parameters
        ----------
        *args, **kwargs: params
            Following list of parameters for the function select_workers.
        """

        workers = self.select_workers(*args, **kwargs)
        ranks = [[self.ranks[w] for w in workers]] * len(workers)
        ranks = self.scatter(ranks, workers=workers, hash=False, broadcast=False)

        # Checking the distribution of the group
        _workers = self.who_has(ranks)
        assert set(workers) == set(
            _workers
        ), """
        Error: Something wrong with scatter. Not all the workers got a piece.
        Expected workers = %s
        Got workers = %s
        """ % (
            workers,
            _workers,
        )

        def _create_comm(ranks):
            comm = default_comm()
            return comm.Create_group(comm.group.Incl(ranks))

        return Comm(self.map(_create_comm, ranks))
